/* RunVCS - Configuration parser
 * Copyright (C) 2009, Daniel Collins <solemnwarning@solemnwarning.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *	* Redistributions of source code must retain the above copyright notice,
 *	  this list of conditions and the following disclaimer.
 *	
 *	* Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 *	
 *	* Neither the name of the developer nor the names of its contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ''AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE DEVELOPER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <string>
#include <vector>

#include "runvcs.hpp"

struct token {
	std::string token;
	
	std::string file;
	int line;
};

typedef std::vector<struct token>::iterator token_iterator;

static void lexer(std::vector<struct token> &tokens, char const *data, std::string file, int line);
static void parser(std::vector<struct token> &tokens);
static void handle_token(token_iterator &token, token_iterator end);
static void tok_profile(token_iterator &token, token_iterator end);
static void tok_allow(token_iterator &token, token_iterator end);

void load_config(std::string file) {
	FILE *fh = fopen(file.c_str(), "r");
	if(!fh) {
		die("Error opening %s: %s", file.c_str(), strerror(errno));
	}
	
	char linebuf[1024];
	int line = 1;
	std::vector<struct token> tokens;
	
	while(fgets(linebuf, 1024, fh)) {
		char *lineptr = linebuf+strspn(linebuf, "\r\n");
		lineptr[strcspn(lineptr, "\r\n")] = '\0';
		
		lexer(tokens, lineptr, file, line);
		line++;
	}
	if(ferror(fh)) {
		die("Error reading %s: %s", file.c_str(), strerror(errno));
	}
	
	fclose(fh);
	
	parser(tokens);
}

#define LEXER_PUSH(t) \
	tokens.push_back((struct token){t,file,line})

static void lexer(std::vector<struct token> &tokens, char const *data, std::string file, int line) {
	char qchar = '\0';
	std::string token = "";
	
	for(; *data; data++) {
		if(*data == '\\') {
			token.append(1, *(++data));
			continue;
		}
		
		if(qchar) {
			if(*data == qchar) {
				LEXER_PUSH(token);
				token.clear();
				
				qchar = '\0';
			}else{
				token.append(1, *data);
			}
			
			continue;
		}
		if(strchr("'\"", *data)) {
			qchar = *data;
			continue;
		}
		
		if(strchr(" \t{}=,", *data)) {
			if(!token.empty()) {
				LEXER_PUSH(token);
				token.clear();
			}
			
			if(strchr("{}=,", *data)) {
				LEXER_PUSH(std::string(1, *data));
			}
		}else{
			token.append(1, *data);
		}
	}
	
	if(qchar) {
		die("%s:%d: Unmatched %c", file.c_str(), line, qchar);
	}
	
	if(!token.empty()) {
		LEXER_PUSH(token);
	}
}

#define CHECK_HAS_TOKEN(e) \
	if(token+1 == end) { \
		die("%s:%d: Expected " e " after '%s'", token->file.c_str(), token->line, (token-1)->token.c_str()); \
	}

#define CHECK_IS_TOKEN(e) \
	CHECK_HAS_TOKEN(e) \
	if(token->token != e) { \
		die("%s:%d: Expected " e " after '%s'", token->file.c_str(), token->line, (token-1)->token.c_str()); \
	} \
	token++;

static void parser(std::vector<struct token> &tokens) {
	token_iterator token = tokens.begin();
	token_iterator end = tokens.end();
	
	while(token != end) {
		handle_token(token, end);
	}
}

static void handle_token(token_iterator &token, token_iterator end) {
	if(token->token == "profile") {
		tok_profile(++token, end);
	}else if(token->token == "allow") {
		tok_allow(++token, end);
	}else{
		die("%s:%d: Unknown directive '%s'", token->file.c_str(), token->line, token->token.c_str());
	}
}

static void tok_profile(token_iterator &token, token_iterator end) {
	std::string name = "";
	int closed = 0, use = 0, deep = 0;
	
	token_iterator begin = token-1;
	CHECK_HAS_TOKEN("profile name");
	name = (token++)->token;
	
	CHECK_IS_TOKEN("{");
	
	std::vector<std::string>::iterator profile = profiles.begin();
	std::vector<std::string>::iterator pend = profiles.end();
	
	while(profile != pend) {
		if(*profile == name) {
			use = 1;
			break;
		}
		
		profile++;
	}
	
	while(token != end && !closed) {
		if(token->token == "}") {
			if(deep) {
				deep--;
			}else{
				closed = 1;
			}
			
			token++;
			continue;
		}
		
		if(use) {
			handle_token(token, end);
			continue;
		}else if(token->token == "{") {
			deep++;
		}
		
		token++;
	}
	
	if(!closed) {
		die("%s:%d: No closing brace", begin->file.c_str(), begin->line);
	}
}

static void tok_allow(token_iterator &token, token_iterator end) {
	token_iterator begin = token-1;
	int closed = 0;
	
	CHECK_IS_TOKEN("svn");
	CHECK_IS_TOKEN("{");
	
	while(token != end && !closed) {
		if(token->token == "}") {
			closed = 1;
			token++;
			
			break;
		}
		
		std::string name = token->token;
		std::string path = token->token;
		
		CHECK_HAS_TOKEN("= , or }");
		token++;
		
		if(token->token == "=") {
			CHECK_HAS_TOKEN("path");
			path = (++token)->token;
			
			CHECK_HAS_TOKEN(", or }");
			token++;
		}
		
		svn_repos.push_back((struct svn_repo){name, path});
		
		if(token->token == "}") {
			continue;
		}else if(token->token != ",") {
			die("%s:%d: Expected , or } after repo", token->file.c_str(), token->line);
		}
		
		token++;
	}
	
	if(!closed) {
		die("%s:%d: No closing brace", begin->file.c_str(), begin->line);
	}
}
